#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'vector_template.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Typed vectors IntVector, DoubleVector and BoolVector.

This module implements a quite simple typed vector which will grow
transparently as needed. The vector is created with a default value,
which will be used for not explicitly set indices.

   vec = IntVector( default_value = 66 )
   vec[0] = 10
   vec[2] = 10

After the 'vec[2] = 10' statement the vector has grown to contain
three elements. The element vec[1] has not been explicitly assigned by
the user, in that case the implementation has 'filled the hole' with
the default value (i.e. 66 in this case). So the statement

   print(vec[1])

will give '66'. The main part of the implementation is in terms of an
"abstract base class" TVector. The TVector class should be not
instantiated directly, instead the child classes IntVector,
DoubleVector or BoolVector should be used.

The C-level has implementations for several fundamental types like
float and size_t not currently implemented in the Python version.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import  sys

from cwrap import CFILE, BaseCClass



class VectorTemplate(BaseCClass):

    def strided_copy(self, slice_range):
        """
        Will create a new copy according to @slice.

        Mainly a support function to support slice based lookup like

           v = IntVector()
           v[0] = 1
           v[1] = 1
           v[100] = 100
           ...
           c = v[0:100:10]

        Now 'c' will be a Intvector() instance containing every tenth
        element from 'v'.
        """
        (start, stop, step) = slice_range.indices(len(self))
        if stop > start:
            return self._strided_copy(start, stop, step)
        else:
            return None

    def __bool__(self):
        """
        Will evaluate to False for empty vector.
        """
        if len(self) == 0:
            return False
        else:
            return True

    def __nonzero__(self):
        return self.__bool__( )


    def __eq__(self, other):
        return self._equal(other)


    def __ne__(self,other):
        return not self.__eq__(other)

    def first_eq(self, other, offset = 0):
        index = self._first_eq(other, offset)
        if index <= -2:
            raise ValueError("Invalid offset")

        return index


    def first_neq(self, other, offset = 0):
        index = self._first_neq(other, offset)
        if index <= -2:
            raise ValueError("Invalid offset")

        return index


    def copy(self):
        """
        Create a new copy of the current vector.
        """
        new = self._alloc_copy( )
        return new

    def __irshift__(self,shift):
        if shift < 0:
            raise ValueError("The shift must be positive")
        self._rshift(shift)
        return self

    def __ilshift__(self,shift):
        if shift < 0:
            raise ValueError("The shift must be positive")
        if shift > len(self):
            raise ValueError("The shift is too large %d > %d" % (shift, len(self)))
        self._lshift( shift)
        return self


    def __rshift__(self,shift):
        copy = self.copy()
        copy >>= shift
        return copy


    def __lshift__(self,shift):
        copy = self.copy()
        copy <<= shift
        return copy

    def __deepcopy__(self, memo):
        new = self.copy()
        return new

    def __init__(self, default_value=0, initial_size=0):
        """
        Creates a new TVector instance.
        """
        c_pointer = self._alloc(initial_size, default_value)
        super(VectorTemplate, self).__init__(c_pointer)
        self.element_size = self._element_size()

    def __contains__(self , value):
        return self._contains(  value)


    def pop(self):
        if len(self) > 0:
            return self._pop()
        else:
            raise ValueError("Trying to pop from empty vector")


    def str_data(self, width, index1, index2, fmt):
        """
        Helper function for str() method.
        """
        data = []
        s = ""
        for index in range(index1, index2):
            data.append(self[index])
        for index in range(len(data)):
            s += fmt % data[index]
            if index % width == (width - 1):
                s += "\n"
        return s


    # The str() method is a verbatim copy of the implementation in
    # ecl_kw.py.
    def str(self, width=5, max_lines=10, fmt=None):
        """
        Return string representation of vector for pretty printing.

        The function will return a string consisting of a header, and
        then a chunk of data. The data will be formatted in @width
        columns, and a maximum of @max_lines lines. If @max_lines is
        not sufficient the first elements in the kewyord are
        represented, a .... continuation line and then the last part
        of the vector. If @max_lines is None all of the vector will be
        printed, irrespectiv of how long it is.

        If a value is given for @fmt that is used as format string for
        each element, otherwise a type-specific default format is
        used. If given the @fmt string should contain spacing between
        the elements. The implementation of the builtin method
        __str__() is based on this method.
        """

        s = ""
        lines = len(self) // width
        if not fmt:
            fmt = self.default_format + " "

        if max_lines is None or lines <= max_lines:
            s += self.str_data(width, 0, len(self), fmt)
        else:
            s1 = width * max_lines // 2
            s += self.str_data(width, 0, s1, fmt)
            s += "   ....   \n"
            s += self.str_data(width, len(self) - s1, len(self), fmt)

        return s

    def __str__(self):
        """
        Returns string representantion of vector.
        """
        return self.str(max_lines=10, width=5)


    def __getitem__(self, index):
        """
        Implements read [] operator - @index can be slice instance.
        """
        if isinstance(index, int):
            length = len(self)
            idx = index
            if idx < 0:
                idx += length

            if 0 <= idx < length:
                return self._iget(idx)
            else:
                raise IndexError('Index must be in range %d <= %d < %d.' % (0, index, length))
        elif isinstance(index, slice):
            return self.strided_copy(index)
        else:
            raise TypeError("Index should be integer or slice type.")

    def __setitem__(self, index, value):
        """
        Implements write [] operator - @index must be integer or slice.
        """
        ls = len(self)
        if isinstance(index, int):
            idx = index
            if idx < 0:
                idx += ls
            self._iset(idx, value)
        elif isinstance( index, slice ):
            for i in range(*index.indices(ls)):
                self[i] = value
        else:
            raise TypeError("Index should be integer type")

    ##################################################################
    # Mathematical operations:

    def __IADD(self, delta, add):
        """
        Low-level function implementing inplace add.

        The __IADD__ function implements the operation:

           v += a

        The variable which is added, i.e. @delta, can either be of the
        same type as the current instance, or a numerical scalar. The
        __IADD__ method implements both add and subtract, based on the
        boolean flag @add.

        The __IADD__ method should not be called directly; but rather
        via the __iadd__, __add__ and __radd__ methods which implement
        the various addition operation, and the corresponding
        subtraction operations: __isub__, __sub__ and __rsub__.
        """
        if type(self) == type(delta):
            if len(self) == len(delta):
                # This is vector + vector operation.
                if not add:
                    delta *= -1
                self._inplace_add(delta)
            else:
                raise ValueError("Incompatible sizes for add self:%d  other:%d" % (len(self), len(delta)))
        else:
            if isinstance(delta, int) or isinstance(delta, float):
                if not add:
                    delta *= -1
                self._shift(delta)
            else:
                raise TypeError("delta has wrong type:%s " % type(delta))

        return self


    def __iadd__(self, delta):
        """
        Implements inplace add. @delta can be vector or scalar.
        """
        return self.__IADD(delta, True)


    def __isub__(self, delta):
        """
        Implements inplace subtract. @delta can be vector or scalar.
        """
        return self.__IADD(delta, False)


    def __radd__(self, delta):
        return self.__add__(delta)


    def __add__(self, delta):
        """
        Implements add operation - creating a new copy.

           b = DoubleVector()
           c = DoubleVector()  // Or alternatively scalar
           ....
           a = b + c
        """
        copy = self._alloc_copy( )
        copy += delta
        return copy

    def __sub__(self, delta):
        """
        Implements subtraction - creating new copy.
        """
        copy = self._alloc_copy( )
        copy -= delta
        return copy


    def __rsub__(self, delta):
        return self.__sub__(delta) * -1


    def __imul__(self, factor):
        """
        Low-level function implementing inplace multiplication.

        The __IMUL__ function implements the operation:

           v *= a

        The variable which is multiplied in, i.e. @factor, can either
        be of the same type as the current instance, or a numerical
        scalar. The __IMUL__ method should not be called directly, but
        rather via the __mul__, __imul__ and __rmul__ functions.
        """

        if type(self) == type(factor):
            # This is vector * vector operation.
            if len(self) == len(factor):
                self._inplace_mul(factor)
            else:
                raise ValueError("Incompatible sizes for mul self:%d  other:%d" % (len(self), len(factor)))
        else:
            if isinstance(factor, int) or isinstance(factor, float):
                self._scale(factor)
            else:
                raise TypeError("factor has wrong type:%s " % type(factor))

        return self


    def __mul__(self, factor):
        copy = self._alloc_copy( )
        copy *= factor
        return copy

    def __rmul__(self, factor):
        return self.__mul__(factor)

    def __div__(self, divisor):
        if isinstance(divisor, int) or isinstance(divisor, float):
            copy = self._alloc_copy()
            copy._div(divisor)
            return copy
        else:
            raise TypeError("Divisor has wrong type:%s" % type(divisor))

    def __truediv__(self, divisor):
        return self.__div__(divisor)

    def __idiv__(self, divisor):
        return self.__div__(divisor)

    def __itruediv__(self, divisor):
        return self.__div__(divisor)

    # End mathematical operations
    #################################################################

    # Essentally implements a = b
    def assign(self, value):
        """
        Implements assignment of type a = b.

        The @value parameter can either be a vector instance, in which
        case the content of @value is copied into the current
        instance, or a scalar in which case all the elements of the
        vector are set to this value:

           v1 = IntVector()
           v2 = IntVector()

           v1[10] = 77
           v2.assign( v1 )      # Now v2 contains identicall content to v1
           ....
           v1.assign( 66 )       # Now v1 is a vector of 11 elements - all equal to 66

        """
        if type(self) == type(value):
            # This is a copy operation
            self._memcpy(value)
        else:
            if isinstance(value, int) or isinstance(value, float):
                self._assign(value)
            else:
                raise TypeError("Value has wrong type")

    def __len__(self):
        """
        The number of elements in the vector.
        """
        return self._size( )


    def printf(self, fmt=None, name=None, stream=sys.stdout):
        """
        See also the str() method which returns string representantion
        of the vector.
        """
        cfile = CFILE(stream)
        if not fmt:
            fmt = self.default_format
        self._fprintf(cfile, name, fmt)


    def max(self):
        if len(self) > 0:
            return self._get_max()
        else:
            raise IndexError("The vector is empty!")

    def min(self):
        if len(self) > 0:
            return self._get_min()
        else:
            raise IndexError("The vector is empty!")


    def minIndex(self, reverse=False):
        """
        @type reverse: bool
        @rtype: int
        """
        if len(self) > 0:
            return self._get_min_index(reverse)
        else:
            raise IndexError("The vector is empty!")

    def maxIndex(self, reverse=False):
        """
        @type reverse: bool
        @rtype: int
        """
        if len(self) > 0:
            return self._get_max_index(reverse)
        else:
            raise IndexError("The vector is empty!")

    def append(self, value):
        self._append(value)

    def deleteBlock(self, index, block_size):
        """
        Remove a block of size @block_size starting at @index.

        After the removal data will be left shifted.
        """
        self._idel_block(index, block_size)

    def sort(self, reverse=False):
        """
        Sort the vector inplace in increasing numerical order or decreasing order if reverse is True.
        @type reverse: bool
        """
        if not reverse:
            self._sort()
        else:
            self._rsort()

    def clear(self):
        self._reset()

    def safeGetByIndex(self, index):
        return self._safe_iget(index)

    def setReadOnly(self, read_only):
        self._set_read_only(read_only)

    def getReadOnly(self):
        return self._get_read_only()

    def setDefault(self, value):
        self._set_default(value)

    def getDefault(self):
        return self._get_default()


    def free(self):
        self._free()

    def __repr__(self):
        return self._create_repr('size = %d' % len(self))

    def permute(self, permutation_vector):
        """
        Reorders this vector based on the indexes in permutation_vector.
        @type permutation_vector: PermutationVector
        """

        self._permute( permutation_vector)

    def permutationSort(self, reverse=False):
        """
        Returns the permutation vector for sorting of this vector. Vector is not sorted.
         @type reverse: bool
         @@rtype: PermutationVector
        """
        if len(self) > 0:
            if not reverse:
                return self._sort_perm()
            else:
                return self._rsort_perm()

        return None


    def asList(self):
        l = [0] * len(self)
        for (index,value) in enumerate(self):
            l[index] = value

        return l

    def selectUnique(self):
        self._select_unique()


    def elementSum(self):
        return self._element_sum( )


    def getDataPtr(self):
        "Low level function which returns a pointer to underlying storage"
        # Observe that the get_data_ptr() function is not implemented
        # for the TimeVector class.
        return self._get_data_ptr()

    def countEqual(self , value):
        return self._count_equal(  value )


    def initRange(self , min_value , max_value , delta):
        """
        Will fill the vector with the values from min_value to
        max_value in steps of delta. The upper limit is guaranteed to
        be inclusive, even if it is not commensurable with the delta.
        """
        if delta == 0:
            raise ValueError("Invalid range")
        else:
            self._init_range(  min_value , max_value , delta )


    @classmethod
    def create_linear(cls, start_value, end_value, num_values):
        vector = cls()
        if not vector._init_linear(start_value, end_value, num_values):
            raise ValueError("init_linear arguments invalid")

        return vector


    @classmethod
    def createRange(cls , min_value , max_value , delta):
        """
        Will create new vector and initialize a range.
        """
        vector = cls( )
        vector.initRange( min_value , max_value , delta )
        return vector

    def _strided_copy(self, *_):
        raise NotImplementedError()
    def _rshift(self, *_):
        raise NotImplementedError()
    def _lshift(self, *_):
        raise NotImplementedError()
    def _alloc(self, *_):
        raise NotImplementedError()
    def _element_size(self, *_):
        raise NotImplementedError()
    def _contains(self, *_):
        raise NotImplementedError()
    def _pop(self, *_):
        raise NotImplementedError()
    def default_format(self, *_):
        raise NotImplementedError()
    def _iget(self, *_):
        raise NotImplementedError()
    def _iset(self, *_):
        raise NotImplementedError()
    def _inplace_add(self, *_):
        raise NotImplementedError()
    def _shift(self, *_):
        raise NotImplementedError()
    def _alloc_copy(self, *_):
        raise NotImplementedError()
    def _inplace_mul(self, *_):
        raise NotImplementedError()
    def _scale(self, *_):
        raise NotImplementedError()
    def _memcpy(self, *_):
        raise NotImplementedError()
    def _assign(self, *_):
        raise NotImplementedError()
    def _size(self, *_):
        raise NotImplementedError()
    def _fprintf(self, *_):
        raise NotImplementedError()
    def _get_max(self, *_):
        raise NotImplementedError()
    def _get_min(self, *_):
        raise NotImplementedError()
    def _get_min_index(self, *_):
        raise NotImplementedError()
    def _get_max_index(self, *_):
        raise NotImplementedError()
    def _append(self, *_):
        raise NotImplementedError()
    def _idel_block(self, *_):
        raise NotImplementedError()
    def _sort(self, *_):
        raise NotImplementedError()
    def _rsort(self, *_):
        raise NotImplementedError()
    def _reset(self, *_):
        raise NotImplementedError()
    def _safe_iget(self, *_):
        raise NotImplementedError()
    def _set_read_only(self, *_):
        raise NotImplementedError()
    def _get_read_only(self, *_):
        raise NotImplementedError()
    def _set_default(self, *_):
        raise NotImplementedError()
    def _get_default(self, *_):
        raise NotImplementedError()
    def _free(self, *_):
        raise NotImplementedError()
    def _permute(self, *_):
        raise NotImplementedError()
    def _sort_perm(self, *_):
        raise NotImplementedError()
    def _rsort_perm(self, *_):
        raise NotImplementedError()
    def _select_unique(self, *_):
        raise NotImplementedError()
    def _element_sum(self, *_):
        raise NotImplementedError()
    def _get_data_ptr(self, *_):
        raise NotImplementedError()
    def _count_equal(self, *_):
        raise NotImplementedError()
    def _init_range(self, *_):
        raise NotImplementedError()



#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
from cwrap import BaseCClass
from ecl.util.util import IntVector
from ecl import EclPrototype
from .cpolyline import CPolyline
from ctypes import c_double

cpair = c_double * 2  # this is a function that maps two doubles to a double*

class GeoRegion(BaseCClass):
    TYPE_NAME = "geo_region"

    _alloc = EclPrototype("void* geo_region_alloc(geo_pointset, bool)", bind=False)
    _free  = EclPrototype("void  geo_region_free(geo_region)")
    _reset = EclPrototype("void  geo_region_reset(geo_region)")
    _get_index_list = EclPrototype("int_vector_ref geo_region_get_index_list(geo_region)")
    _select_inside_polygon    = EclPrototype("void geo_region_select_inside_polygon(geo_region, geo_polygon)")
    _select_outside_polygon   = EclPrototype("void geo_region_select_outside_polygon(geo_region, geo_polygon)")
    _deselect_inside_polygon  = EclPrototype("void geo_region_deselect_inside_polygon(geo_region, geo_polygon)")
    _deselect_outside_polygon = EclPrototype("void geo_region_deselect_outside_polygon(geo_region, geo_polygon)")
    _select_above_line        = EclPrototype("void geo_region_select_above_line(geo_region, double*, double*)")
    _select_below_line        = EclPrototype("void geo_region_select_below_line(geo_region, double*, double*)")
    _deselect_above_line      = EclPrototype("void geo_region_deselect_above_line(geo_region, double*, double*)")
    _deselect_below_line      = EclPrototype("void geo_region_deselect_below_line(geo_region, double*, double*)")


    def __init__(self, pointset, preselect=False):
        self._preselect = True if preselect else False
        c_ptr = self._alloc(pointset, self._preselect)
        if c_ptr:
            super(GeoRegion, self).__init__(c_ptr)
        else:
            raise ValueError('Could not construct GeoRegion from pointset %s.' % pointset)


    def getActiveList(self):
        return self._get_index_list()

    def _assert_polygon(self, polygon):
        if not isinstance(polygon, CPolyline):
            raise ValueError('Need to select with a CPolyline, not %s.'
                             % type(polygon))


    def _construct_cline(self, line):
        """Takes a line ((x1,y1), (x2,y2)) and returns two double[2]* but
        reordered to (x1x2, y1y2).
        """
        try:
            p1, p2 = line
            x1, y1 = map(float, p1)
            x2, y2 = map(float, p2)
        except Exception as err:
            err_msg = 'Select with pair ((x1,y1), (x2,y2)), not %s (%s).'
            raise ValueError(err_msg % (line, err))
        x1x2_ptr = cpair(x1, x2)
        y1y2_ptr = cpair(y1, y2)
        return x1x2_ptr, y1y2_ptr


    def select_inside(self, polygon):
        self._assert_polygon(polygon)
        self._select_inside_polygon(polygon)

    def select_outside(self, polygon):
        self._assert_polygon(polygon)
        self._select_outside_polygon(polygon)

    def deselect_inside(self, polygon):
        self._assert_polygon(polygon)
        self._deselect_inside_polygon(polygon)

    def deselect_outside(self, polygon):
        self._assert_polygon(polygon)
        self._deselect_outside_polygon(polygon)


    def select_above(self, line):
        x_ptr, y_ptr = self._construct_cline(line)
        self._select_above_line(x_ptr, y_ptr)

    def select_below(self, line):
        x_ptr, y_ptr = self._construct_cline(line)
        self._select_below_line(x_ptr, y_ptr)

    def deselect_above(self, line):
        x_ptr, y_ptr = self._construct_cline(line)
        self._deselect_above_line(x_ptr, y_ptr)

    def deselect_below(self, line):
        x_ptr, y_ptr = self._construct_cline(line)
        self._deselect_below_line(x_ptr, y_ptr)


    def __len__(self):
        """Returns the size of the active list, not the size of the
        underlying pointset"""
        return len(self._get_index_list())

    def __repr__(self):
        ls = len(self)
        il = repr(self.getActiveList())
        pres = 'preselected' if self._preselect else 'not preselected'
        return self._create_repr('size=%d, active_list=<%s>, %s' % (ls, il, pres))

    def free(self):
        self._free()



import inspect

def debug_msg(msg):
    record = inspect.stack()[1]
    frame = record[0]
    info = inspect.getframeinfo( frame )
    return "FILE: %s  LINE: %s  Msg: %s" % (info.filename, info.lineno, msg)



from ecl.util.test import TestRun
from ecl.util.test import path_exists
from ecl.util.test import ExtendedTestCase
from ecl.util.test import SourceEnumerator
from ecl.util.test import TestArea , TestAreaContext
from ecl.util.test import TempArea , TempAreaContext
from ecl.util.test import ErtTestRunner
from ecl.util.test import PathContext
from ecl.util.test import LintTestCase
from ecl.util.test import ImportTestCase



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_rft_cell.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from cwrap import BaseCClass
from ecl import EclPrototype


class RFTCell(BaseCClass):
    """The RFTCell is a base class for the cells which are part of an RFT/PLT.

    The RFTCell class contains the elements which are common to both
    RFT and PLT. The list of common elements include the corrdinates
    (i,j,k) the pressure and the depth of the cell. Actual user access
    should be based on the derived classes EclRFTCell and EclPLTCell.

    Observe that from june 2013 the properties i,j and k which return
    offset 1 coordinate values are deprecated, and you should rather
    use the methods get_i(), get_j() and get_k() which return offset 0
    coordinate values.
    """
    TYPE_NAME = "rft_cell"
    _free         = EclPrototype("void ecl_rft_cell_free(rft_cell)")
    _get_pressure = EclPrototype("double ecl_rft_cell_get_pressure(rft_cell)")
    _get_depth    = EclPrototype("double ecl_rft_cell_get_depth(rft_cell)")
    _get_i        = EclPrototype("int ecl_rft_cell_get_i(rft_cell)")
    _get_j        = EclPrototype("int ecl_rft_cell_get_j(rft_cell)")
    _get_k        = EclPrototype("int ecl_rft_cell_get_k(rft_cell)")


    def free(self):
        self._free()

    def get_i(self):
        return self._get_i()

    def get_j(self):
        return self._get_j()

    def get_k(self):
        return self._get_k()

    def get_ijk(self):
        return (self.get_i(), self.get_j(), self.get_k())

    @property
    def pressure(self):
        return self._get_pressure()

    @property
    def depth(self):
        return self._get_depth()


#################################################################


class EclRFTCell(RFTCell):
    TYPE_NAME  = "ecl_rft_cell"
    _alloc_RFT = EclPrototype("void* ecl_rft_cell_alloc_RFT(int, int, int, double, double, double, double)", bind = False)
    _get_swat  = EclPrototype("double ecl_rft_cell_get_swat(ecl_rft_cell)")
    _get_soil  = EclPrototype("double ecl_rft_cell_get_soil(ecl_rft_cell)")
    _get_sgas  = EclPrototype("double ecl_rft_cell_get_sgas(ecl_rft_cell)")

    def __init__(self, i, j, k, depth, pressure, swat, sgas):
        c_ptr = self._alloc_RFT(i, j, k, depth, pressure, swat, sgas)
        super(EclRFTCell, self).__init__(c_ptr)

    @property
    def swat(self):
        return self._get_swat()

    @property
    def sgas(self):
        return self._get_sgas()

    @property
    def soil(self):
        return 1 - (self._get_sgas() + self._get_swat())


#################################################################


class EclPLTCell(RFTCell):
    TYPE_NAME = "ecl_plt_cell"
    _alloc_PLT = EclPrototype("void* ecl_rft_cell_alloc_PLT(int, int, int, double, double, double, double, double, double, double, double, double, double, double)", bind=False)
    _get_orat = EclPrototype("double ecl_rft_cell_get_orat(ecl_plt_cell)")
    _get_grat = EclPrototype("double ecl_rft_cell_get_grat(ecl_plt_cell)")
    _get_wrat = EclPrototype("double ecl_rft_cell_get_wrat(ecl_plt_cell)")

    _get_flowrate = EclPrototype("double ecl_rft_cell_get_flowrate(ecl_plt_cell)")
    _get_oil_flowrate = EclPrototype("double ecl_rft_cell_get_oil_flowrate(ecl_plt_cell)")
    _get_gas_flowrate = EclPrototype("double ecl_rft_cell_get_gas_flowrate(ecl_plt_cell)")
    _get_water_flowrate = EclPrototype("double ecl_rft_cell_get_water_flowrate(ecl_plt_cell)")

    _get_conn_start = EclPrototype("double ecl_rft_cell_get_connection_start(ecl_plt_cell)")
    _get_conn_end   = EclPrototype("double ecl_rft_cell_get_connection_end(ecl_plt_cell)")


    def __init__(self, i, j, k, depth, pressure, orat, grat, wrat, conn_start,
                 conn_end, flowrate, oil_flowrate, gas_flowrate, water_flowrate):
        c_ptr = self._alloc_PLT(i, j, k, depth, pressure, orat, grat, wrat,
                                conn_start, conn_end, flowrate, oil_flowrate,
                                gas_flowrate, water_flowrate)
        super(EclPLTCell, self).__init__(c_ptr)


    @property
    def orat(self):
        return self._get_orat()

    @property
    def grat(self):
        return self._get_grat()

    @property
    def wrat(self):
        return self._get_wrat()

    @property
    def conn_start(self):
        """Will return the length from wellhead(?) to connection.

        For MSW wells this property will return the distance from a
        fixed point (wellhead) to the current connection. This value
        will be used to sort the completed cells along the well
        path. In the case of non MSW wells this will just return a
        fixed default value.
        """
        return self._get_conn_start()

    @property
    def conn_end(self):
        """Will return the length from wellhead(?) to connection end.

        For MSW wells this property will return the distance from a
        fixed point (wellhead) to the current connection end. This value
        will be used to sort the completed cells along the well
        path. In the case of non MSW wells this will just return a
        fixed default value.
        """
        return self._get_conn_end()

    @property
    def flowrate(self):
        return self._get_flowrate()

    @property
    def oil_flowrate(self):
        return self._get_oil_flowrate()

    @property
    def gas_flowrate(self):
        return self._get_gas_flowrate()

    @property
    def water_flowrate(self):
        return self._get_water_flowrate()



from ecl import EclFileEnum, EclFileFlagEnum, EclPhaseEnum, EclUnitTypeEnum , EclUtil
from ecl import EclTypeEnum, EclDataType
from ecl.summary import EclSumVarType
from ecl.summary import EclSumTStep
from ecl.summary import EclSum #, EclSumVector, EclSumNode, EclSMSPECNode
from ecl.summary import EclSumKeyWordVector
from ecl.rft import EclPLTCell, EclRFTCell
from ecl.rft import EclRFT, EclRFTFile
from ecl.eclfile import FortIO, openFortIO
from ecl.eclfile import EclKW
from ecl.eclfile import Ecl3DKW
from ecl.eclfile import EclFileView
from ecl.eclfile import EclFile , openEclFile
from ecl.eclfile import Ecl3DFile
from ecl.eclfile import EclInitFile
from ecl.eclfile import EclRestartFile
from ecl.grid import EclGrid
from ecl.grid import EclRegion
from ecl.gravimetry import EclSubsidence
from ecl.gravimetry import phase_deltag, deltag
from ecl.gravimetry import EclGrav
from ecl.summary import EclSumNode
from ecl.summary import EclSumVector
from ecl.summary import EclNPV , NPVPriceVector
from ecl.summary import EclCmp
from ecl.grid import EclGridGenerator



import numbers
import os
import os.path
import traceback
import sys

try:
    from unittest2 import TestCase
except ImportError:
    from unittest import TestCase

from .source_enumerator import SourceEnumerator
from ecl.util.util import installAbortSignals
from ecl.util.util import Version


# Function wrapper which can be used to add decorator @log_test to test
# methods. When a test has been decorated with @log_test it will print
# "starting: <test_name>" when a method is complete and "complete: <test_name>"
# when the method is complete. Convenient when debugging tests which fail hard
# or lock up.

def log_test(test):
    def wrapper(*args):
        sys.stderr.write("starting: %s \n" % test.__name__)
        test(*args)
        sys.stderr.write("complete: %s \n" % test.__name__)
    return wrapper



class _AssertNotRaisesContext(object):

    def __init__(self, test_class):
        super(_AssertNotRaisesContext, self).__init__()
        self._test_class = test_class

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            try:
                exc_name = exc_type.__name__
            except AttributeError:
                exc_name = str(exc_type)
            self._test_class.fail("Exception: %s raised\n%s" % (exc_name, traceback.print_exception(exc_type, exc_value, tb)))
        return True


"""
This class provides some extra functionality for testing values that are almost equal.
"""
class ExtendedTestCase(TestCase):
    TESTDATA_ROOT = None
    SHARE_ROOT = None
    SOURCE_ROOT = None


    def __init__(self , *args , **kwargs):
        installAbortSignals()
        super(ExtendedTestCase , self).__init__(*args , **kwargs)


    def assertFloatEqual(self, first, second, msg=None, tolerance=1e-6):
        try:
            f_first, f_second = float(first), float(second)
            diff = abs(f_first - f_second)
            scale = max(1, abs(first) + abs(second))
            if msg is None:
                msg = "Floats not equal: |%f - %f| > %g" % (f_first, f_second, tolerance)
            self.assertTrue(diff < tolerance * scale, msg=msg)
        except TypeError:
            self.fail("Cannot compare as floats: %s (%s) and %s (%s)" %
                      (first, type(first), second, type(second)))


    def assertAlmostEqualList(self, first, second, msg=None, tolerance=1e-6):
        if len(first) != len(second):
            self.fail("Lists are not of same length!")

        for index in range(len(first)):
            self.assertFloatEqual(
                    first[index], second[index],
                    msg=msg, tolerance=tolerance
                    )


    def assertImportable(self, module_name):
        try:
            __import__(module_name)
        except ImportError:
            tb = traceback.format_exc()
            self.fail("Module %s not found!\n\nTrace:\n%s" % (module_name, str(tb)))
        except Exception:
            tb = traceback.format_exc()
            self.fail("Import of module %s caused errors!\n\nTrace:\n%s" % (module_name, str(tb)))


    def assertFilesAreEqual(self, first, second):
        if not self.__filesAreEqual(first, second):
            self.fail("Buffer contents of files are not identical!")


    def assertFilesAreNotEqual(self, first, second):
        if self.__filesAreEqual(first, second):
            self.fail("Buffer contents of files are identical!")

    def assertFileExists(self, path):
        if not os.path.exists(path) or not os.path.isfile(path):
            self.fail("The file: %s does not exist!" % path)

    def assertDirectoryExists(self, path):
        if not os.path.exists(path) or not os.path.isdir(path):
            self.fail("The directory: %s does not exist!" % path)

    def assertFileDoesNotExist(self, path):
        if os.path.exists(path) and os.path.isfile(path):
            self.fail("The file: %s exists!" % path)

    def assertDirectoryDoesNotExist(self, path):
        if os.path.exists(path) and os.path.isdir(path):
            self.fail("The directory: %s exists!" % path)

    def __filesAreEqual(self, first, second):
        buffer1 = open(first, "rb").read()
        buffer2 = open(second, "rb").read()

        return buffer1 == buffer2

    def assertEnumIsFullyDefined(self, enum_class, enum_name, source_path, verbose=False):
        if self.SOURCE_ROOT is None:
            raise Exception("SOURCE_ROOT is not set.")

        enum_values = SourceEnumerator.findEnumerators(enum_name, os.path.join( self.SOURCE_ROOT , source_path))

        for identifier, value in enum_values:
            if verbose:
                print("%s = %d" % (identifier, value))

            self.assertTrue(identifier in enum_class.__dict__, "Enum does not have identifier: %s" % identifier)
            class_value = enum_class.__dict__[identifier]
            self.assertEqual(class_value, value, "Enum value for identifier: %s does not match: %s != %s" % (identifier, class_value, value))


    @classmethod
    def createSharePath(cls, path):
        if cls.SHARE_ROOT is None:
            raise Exception("Trying to create directory rooted in 'SHARE_ROOT' - variable 'SHARE_ROOT' is not set.")
        return os.path.realpath(os.path.join(cls.SHARE_ROOT , path))


    @classmethod
    def createTestPath(cls, path):
        if cls.TESTDATA_ROOT is None:
            raise Exception("Trying to create directory rooted in 'TESTDATA_ROOT' - variable 'TESTDATA_ROOT' has not been set.")
        return os.path.realpath(os.path.join(cls.TESTDATA_ROOT , path))


    def assertNotRaises(self, func=None):

        context = _AssertNotRaisesContext(self)
        if func is None:
            return context

        with context:
            func()

    @staticmethod
    def slowTestShouldNotRun():
        """
        @param: The slow test flag can be set by environment variable SKIP_SLOW_TESTS = [True|False]
        """

        return os.environ.get("SKIP_SLOW_TESTS", "False") == "True"


    @staticmethod
    def requireVersion(major , minor , micro = "git"):
        required_version = Version(major, minor , micro)
        current_version = Version.currentVersion()

        if required_version < current_version:
            return True
        else:
            return False



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_subsidence.py' is part of ERT - Ensemble based
#  Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Calculate dynamic change in gravitational strength.

The ecl_subsidence module contains functionality to load time-lapse ECLIPSE
results and calculate the change in seafloor subsidence between the
different surveys. The implementation is a thin wrapper around the
ecl_subsidence.c implementation in the libecl library.
"""
from cwrap import BaseCClass
from ecl import EclPrototype
from ecl.util.util import monkey_the_camel
import ecl.grid

class EclSubsidence(BaseCClass):
    """
    Holding ECLIPSE results for calculating subsidence changes.

    The EclSubsidence class is a collection class holding the results from
    ECLIPSE forward modelling of subsidence surveys. Observe that the
    class is focused on the ECLIPSE side of things, and does not have
    any notion of observed values or measurement locations; that
    should be handled by the scope using the EclSubsidence class.

    Typical use of the EclSubsidence class involves the following steps:

      1. Create the EclSubsidence instance.
      2. Add surveys with the add_survey_XXXX() methods.
      3. Evalute the subsidence response with the eval() method.
    """
    TYPE_NAME = "ecl_subsidence"
    _alloc               = EclPrototype("void* ecl_subsidence_alloc( ecl_grid , ecl_file )" , bind = False)
    _free                = EclPrototype("void ecl_subsidence_free( ecl_subsidence )")
    _add_survey_PRESSURE = EclPrototype("void*  ecl_subsidence_add_survey_PRESSURE( ecl_subsidence , char* , ecl_file_view )")
    _eval                = EclPrototype("double ecl_subsidence_eval( ecl_subsidence , char* , char* , ecl_region , double , double , double, double, double)")
    _eval_geertsma       = EclPrototype("double ecl_subsidence_eval_geertsma( ecl_subsidence , char* , char* , ecl_region , double , double , double, double, double, double)")
    _has_survey          = EclPrototype("bool  ecl_subsidence_has_survey( ecl_subsidence , char*)")

    def __init__( self, grid, init_file ):
        """
        Creates a new EclSubsidence instance.

        The input arguments @grid and @init_file should be instances
        of EclGrid and EclFile respectively.
        """
        self.init_file = init_file   # Inhibit premature garbage collection of init_file
        c_ptr = self._alloc( grid , init_file )
        super( EclSubsidence , self ).__init__( c_ptr )


    def __contains__(self , survey_name):
        return self._has_survey( survey_name )



    def add_survey_PRESSURE( self, survey_name, restart_file ):
        """
        Add new survey based on PRESSURE keyword.

        Add a new survey; in this context a survey is the state of
        reservoir, i.e. an ECLIPSE restart file. The @survey_name
        input argument will be used when refering to this survey at a
        later stage. The @restart_file input argument should be an
        EclFile instance with data from one report step. A typical way
        to load the @restart_file argument is:

           import datetime
           import ecl.ecl.ecl as ecl
           ...
           ...
           date = datetime.datetime( year , month , day )
           restart_file1 = ecl.EclFile.restart_block( "ECLIPSE.UNRST" , dtime = date)
           restart_file2 = ecl.EclFile.restart_block( "ECLIPSE.UNRST" , report_step = 67 )

        The pore volume is calculated from the initial pore volume and
        the PRESSURE keyword from the restart file.
        """
        self._add_survey_PRESSURE( survey_name, restart_file)


    def eval_geertsma(self, base_survey, monitor_survey, pos, youngs_modulus, poisson_ratio, seabed, region=None):
        if not base_survey in self:
            raise KeyError("No such survey: %s" % base_survey)

        if monitor_survey is not None:
            if not monitor_survey in self:
                raise KeyError("No such survey: %s" % monitor_survey)

        return self._eval_geertsma(base_survey, monitor_survey, region, pos[0], pos[1], pos[2], youngs_modulus, poisson_ratio, seabed)

    def eval(self, base_survey, monitor_survey, pos, compressibility, poisson_ratio, region=None):
        """
        Calculates the subsidence change between two surveys.

        This is the method everything is leading up to; will calculate
        the change in subsidence, in centimeters,
        between the two surveys named @base_survey and
        @monitor_survey.

        The monitor survey can be 'None' - the resulting answer has
        nothing whatsovever to do with subsidence, but can be
        interesting to determine the numerical size of the quantities
        which are subtracted in a 4D study.

        The @pos argument should be a tuple of three elements with the
        (utm_x , utm_y , depth) position where we want to evaluate the
        change in subsidence.

        If supplied the optional argument @region should be an
        EclRegion() instance; this region will be used to limit the
        part of the reserviour included in the subsidence calculations.

        The argument @compressibility is the total reservoir compressibility.
        """
        if not base_survey in self:
            raise KeyError("No such survey: %s" % base_survey)

        if not monitor_survey in self:
            raise KeyError("No such survey: %s" % monitor_survey)

        return self._eval(base_survey, monitor_survey, region, pos[0], pos[1], pos[2], compressibility,poisson_ratio)



    def free(self):
        self._free( )


monkey_the_camel(EclSubsidence, 'evalGeertsma', EclSubsidence.eval_geertsma)



from cwrap import BaseCEnum

class WellTypeEnum(BaseCEnum):
    TYPE_NAME = "well_type_enum"
    ECL_WELL_ZERO = None
    ECL_WELL_PRODUCER = None
    ECL_WELL_WATER_INJECTOR = None
    ECL_WELL_GAS_INJECTOR = None
    ECL_WELL_OIL_INJECTOR = None

WellTypeEnum.addEnum("ECL_WELL_ZERO", 0)
WellTypeEnum.addEnum("ECL_WELL_PRODUCER", 1)
WellTypeEnum.addEnum("ECL_WELL_OIL_INJECTOR", 2)
WellTypeEnum.addEnum("ECL_WELL_WATER_INJECTOR", 3)
WellTypeEnum.addEnum("ECL_WELL_GAS_INJECTOR", 4)




#  Copyright (C) 2016  Statoil ASA, Norway.
#
#  The file 'ecl_sum_var_type.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from cwrap import BaseCEnum



class EclSumVarType(BaseCEnum):
    TYPE_NAME = "ecl_sum_var_type"
    ECL_SMSPEC_INVALID_VAR            = None
    ECL_SMSPEC_FIELD_VAR              = None
    ECL_SMSPEC_REGION_VAR             = None
    ECL_SMSPEC_GROUP_VAR              = None
    ECL_SMSPEC_WELL_VAR               = None
    ECL_SMSPEC_SEGMENT_VAR            = None
    ECL_SMSPEC_BLOCK_VAR              = None
    ECL_SMSPEC_AQUIFER_VAR            = None
    ECL_SMSPEC_COMPLETION_VAR         = None
    ECL_SMSPEC_NETWORK_VAR            = None
    ECL_SMSPEC_REGION_2_REGION_VAR    = None
    ECL_SMSPEC_LOCAL_BLOCK_VAR        = None
    ECL_SMSPEC_LOCAL_COMPLETION_VAR   = None
    ECL_SMSPEC_LOCAL_WELL_VAR         = None
    ECL_SMSPEC_MISC_VAR               = None


EclSumVarType.addEnum("ECL_SMSPEC_INVALID_VAR", 0)
EclSumVarType.addEnum("ECL_SMSPEC_FIELD_VAR", 1)
EclSumVarType.addEnum("ECL_SMSPEC_REGION_VAR", 2)
EclSumVarType.addEnum("ECL_SMSPEC_GROUP_VAR", 3)
EclSumVarType.addEnum("ECL_SMSPEC_WELL_VAR", 4)
EclSumVarType.addEnum("ECL_SMSPEC_SEGMENT_VAR", 5)
EclSumVarType.addEnum("ECL_SMSPEC_BLOCK_VAR", 6)
EclSumVarType.addEnum("ECL_SMSPEC_AQUIFER_VAR", 7)
EclSumVarType.addEnum("ECL_SMSPEC_COMPLETION_VAR", 8)
EclSumVarType.addEnum("ECL_SMSPEC_NETWORK_VAR", 9)
EclSumVarType.addEnum("ECL_SMSPEC_REGION_2_REGION_VAR", 10)
EclSumVarType.addEnum("ECL_SMSPEC_LOCAL_BLOCK_VAR", 11)
EclSumVarType.addEnum("ECL_SMSPEC_LOCAL_COMPLETION_VAR", 12)
EclSumVarType.addEnum("ECL_SMSPEC_LOCAL_WELL_VAR", 13)
EclSumVarType.addEnum("ECL_SMSPEC_MISC_VAR", 14)
    




#  Copyright (C) 2013  Statoil ASA, Norway.
#
#  The file 'test_work_area.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
import os.path

from cwrap import BaseCClass
from ert.enkf import EnKFMain, EnkfPrototype

class ErtTest(BaseCClass):
    TYPE_NAME = "ert_test"

    _alloc         = EnkfPrototype("void* ert_test_context_alloc_python( char* , char*)", bind = False)
    _set_store     = EnkfPrototype("void* ert_test_context_set_store( ert_test , bool)")
    _free          = EnkfPrototype("void  ert_test_context_free( ert_test )")
    _get_cwd       = EnkfPrototype("char* ert_test_context_get_cwd( ert_test )")
    _get_enkf_main = EnkfPrototype("enkf_main_ref ert_test_context_get_main( ert_test )")



    def __init__(self, test_name, model_config, store_area=False):
        if not os.path.exists(model_config):
            raise IOError("The configuration file: %s does not exist" % model_config)
        else:
            c_ptr = self._alloc(test_name, model_config)
            super(ErtTest, self).__init__(c_ptr)
            self.setStore(store_area)

        self.__ert = None

    def setStore(self, store):
        self._set_store(store)

    def getErt(self):
        """ @rtype: EnKFMain """
        if self.__ert is None:
            self.__ert = self._get_enkf_main()

        return self.__ert

    def free(self):
        ert = self.getErt()
        ert.umount()
        self._free()

    def installWorkflowJob(self, job_name, job_path):
        """ @rtype: bool """
        if os.path.exists(job_path) and os.path.isfile(job_path):
            ert = self.getErt()
            workflow_list = ert.getWorkflowList()

            workflow_list.addJob(job_name, job_path)
            return workflow_list.hasJob(job_name)
        else:
            return False

    def runWorkflowJob(self, job_name, *arguments):
        """ @rtype: bool """
        ert = self.getErt()
        workflow_list = ert.getWorkflowList()

        if workflow_list.hasJob(job_name):
            job = workflow_list.getJob(job_name)
            job.run(ert, [arg for arg in arguments])
            return True
        else:
            return False


    def getCwd(self):
        """
        Returns the current working directory of this context.
        @rtype: string
        """
        return self._get_cwd()



class ErtTestContext(object):
    def __init__(self, test_name, model_config, store_area=False):
        self.__test_name = test_name
        self.__model_config = model_config
        self.__store_area = store_area
        self.__test_context = ErtTest(self.__test_name, self.__model_config, store_area=self.__store_area)


    def __enter__(self):
        """ @rtype: ErtTest """
        return self.__test_context


    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.__test_context
        return False


    def getErt(self):
        return self.__test_context.getErt()


    def getCwd(self):
        """
        Returns the current working directory of this context.
        @rtype: string
        """
        return self.__test_context.getCwd()



#  Copyright (C) 2015  Statoil ASA, Norway.
#
#  The file 'well_trajectory.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

import sys
from os.path import isfile
from collections import namedtuple

TrajectoryPoint = namedtuple('TrajectoryPoint',
                             ['utm_x', 'utm_y',
                              'measured_depth', 'true_vertical_depth',
                              'zone'])

def _read_point(line):
    line = line.partition("--")[0]
    line = line.strip()
    if not line:
        return None

    point = line.split()
    if len(point) not in (4, 5):
        fmt = 'utm_x utm_y md tvd [zone]'
        err = 'Trajectory data file not on correct format: "%s".'
        err += '  zone is optional.'
        raise UserWarning(err % fmt)
    return point

def _parse_point(point):
    try:
        utm_x = float(point[0])
        utm_y = float(point[1])
        md = float(point[2])
        tvd = float(point[3])
    except ValueError:
        raise UserWarning("Error: Failed to extract data from line %s\n" % str(point))
    zone = None
    if len(point) > 4:
        zone = point[4]
    return TrajectoryPoint(utm_x, utm_y, md, tvd, zone)


class WellTrajectory:

    def __init__(self, filename):
        self._points = []
        if not isfile(filename):
            raise IOError('No such file "%s"' % filename)

        with open(filename) as fileH:
            for line in fileH.readlines():
                point = _read_point(line)
                if not point:
                    continue
                traj = _parse_point(point)
                self._points.append(traj)

    def __len__(self):
        return len(self._points)


    def __getitem__(self, index):
        if index < 0:
            index += len(self)

        return self._points[index]


    def __repr__(self):
        return 'WellTrajectory(len=%d)' % len(self)

    def __str__(self):
        return 'WellTrajectory(%s)' % str(self._points)



#  Copyright (C) 2015  Statoil ASA, Norway.
#
#  The file 'cthread_pool.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

import ctypes

from cwrap import BaseCClass
from ecl import EclPrototype

import weakref


class CThreadPool(BaseCClass):
    TYPE_NAME = "thread_pool"

    _alloc   = EclPrototype("void* thread_pool_alloc(int, bool)", bind = False)
    _free    = EclPrototype("void thread_pool_free(thread_pool)")
    _add_job = EclPrototype("void thread_pool_add_job(thread_pool, void*, void*)")
    _join    = EclPrototype("void thread_pool_join(thread_pool)")

    def __init__(self, pool_size, start=True):
        c_ptr = self._alloc(pool_size, start)
        super(CThreadPool, self).__init__(c_ptr)
        self.arg_list = []

    def addTaskFunction(self, name, lib, c_function_name):
        function = CThreadPool.lookupCFunction(lib, c_function_name)
        self_ref = weakref.ref(self)  # avoid circular dependencies

        def wrappedFunction(arg):
            return self_ref().addTask(function, arg)

        setattr(self, name, wrappedFunction)

    def addTask(self, cfunc, arg):
        """
        The function should come from CThreadPool.lookupCFunction().
        """
        if isinstance(arg, BaseCClass):
            arg_ptr = BaseCClass.from_param(arg)
        else:
            arg_ptr = arg

        self.arg_list.append(arg)
        self._add_job(cfunc, arg_ptr)

    def join(self):
        self._join()

    def free(self):
        self.join()
        self._free()

    @staticmethod
    def lookupCFunction(lib, name):
        if isinstance(lib, ctypes.CDLL):
            func = getattr(lib, name)
            return func
        else:
            raise TypeError("The lib argument must be of type ctypes.CDLL")


class CThreadPoolContextManager(object):
    def __init__(self, tp):
        self.__tp = tp

    def __enter__(self):
        return self.__tp

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__tp.join()
        return False


def startCThreadPool(size):
    return CThreadPoolContextManager(CThreadPool(size, start=True))



#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'fault_block_layer.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from __future__ import print_function
from cwrap import BaseCClass

from ecl.util.util import monkey_the_camel
from ecl import EclDataType
from ecl import EclPrototype
from ecl.grid.faults import Fault

class FaultBlockLayer(BaseCClass):
    TYPE_NAME = "fault_block_layer"
    _alloc                = EclPrototype("void*           fault_block_layer_alloc(ecl_grid,  int)", bind = False)
    _free                 = EclPrototype("void            fault_block_layer_free(fault_block_layer)")
    _size                 = EclPrototype("int             fault_block_layer_get_size(fault_block_layer)")
    _iget_block           = EclPrototype("fault_block_ref fault_block_layer_iget_block(fault_block_layer, int)")
    _add_block            = EclPrototype("fault_block_ref fault_block_layer_add_block(fault_block_layer, int)")
    _get_block            = EclPrototype("fault_block_ref fault_block_layer_get_block(fault_block_layer, int)")
    _del_block            = EclPrototype("void            fault_block_layer_del_block(fault_block_layer, int)")
    _has_block            = EclPrototype("bool            fault_block_layer_has_block(fault_block_layer, int)")
    _scan_keyword         = EclPrototype("bool            fault_block_layer_scan_kw(fault_block_layer, ecl_kw)")
    _load_keyword         = EclPrototype("bool            fault_block_layer_load_kw(fault_block_layer, ecl_kw)")
    _getK                 = EclPrototype("int             fault_block_layer_get_k(fault_block_layer)")
    _get_next_id          = EclPrototype("int             fault_block_layer_get_next_id(fault_block_layer)")
    _scan_layer           = EclPrototype("void            fault_block_layer_scan_layer(fault_block_layer, layer)")
    _insert_block_content = EclPrototype("void            fault_block_layer_insert_block_content(fault_block_layer, fault_block)")
    _export_kw            = EclPrototype("bool            fault_block_layer_export(fault_block_layer, ecl_kw)")
    _get_layer            = EclPrototype("layer_ref       fault_block_layer_get_layer(fault_block_layer)")


    def __init__(self, grid, k):
        c_ptr = self._alloc(grid, k)
        if c_ptr:
            super(FaultBlockLayer, self).__init__(c_ptr)
        else:
            raise ValueError("Invalid input - failed to create FaultBlockLayer")

        # The underlying C implementation uses lazy evaluation and
        # needs to hold on to the grid reference. We therefore take
        # references to it here, to protect against premature garbage
        # collection.
        self.grid_ref = grid


    def __len__(self):
        return self._size()


    def __repr__(self):
        return self._create_repr('size=%d, k=%d' % (len(self), self.get_k()))

    def __getitem__(self, index):
        """
        @rtype: FaultBlock
        """
        if isinstance(index, int):
            if index < 0:
                index += len(self)

            if 0 <= index < len(self):
                return self._iget_block(index).setParent(self)
            else:
                raise IndexError("Index:%d out of range: [0,%d)" % (index, len(self)))
        elif isinstance(index,tuple):
            i,j = index
            if 0 <= i < self.grid_ref.getNX() and 0 <= j < self.grid_ref.getNY():
                geo_layer = self.getGeoLayer()
                block_id = geo_layer[i,j]
                if block_id == 0:
                    raise ValueError("No fault block defined for location (%d,%d)" % (i,j))
                else:
                    return self.getBlock(block_id)
            else:
                raise IndexError("Invalid i,j : (%d,%d)" % (i,j))
        else:
            raise TypeError("Index should be integer type")

    def __contains__(self, block_id):
        return self._has_block(block_id)


    def scan_keyword(self, fault_block_kw):
        """
        Will reorder the block ids, and ensure single connectedness. Assign block_id to zero blocks.
        """
        ok = self._scan_keyword(fault_block_kw)
        if not ok:
            raise ValueError("The fault block keyword had wrong type/size: type:%s  size:%d  grid_size:%d" % (fault_block_kw.type_name, len(fault_block_kw), self.grid_ref.getGlobalSize()))


    def load_keyword(self, fault_block_kw):
        """
        Will load directly from keyword - without reorder; ignoring zero.
        """
        ok = self._load_keyword(fault_block_kw)
        if not ok:
            raise ValueError("The fault block keyword had wrong type/size:  type:%s  size:%d  grid_size:%d" % (fault_block_kw.typeName(), len(fault_block_kw), self.grid_ref.getGlobalSize()))


    def get_block(self, block_id):
        """
        @rtype: FaultBlock
        """
        if block_id in self:
            return self._get_block(block_id).setParent(self)
        else:
            raise KeyError("No blocks with ID:%d in this layer" % block_id)


    def delete_block(self, block_id):
        if block_id in self:
            self._del_block(block_id)
        else:
            raise KeyError("No blocks with ID:%d in this layer" % block_id)

    def add_block(self, block_id=None):
        if block_id is None:
            block_id = self.getNextID()

        if block_id in self:
            raise KeyError("Layer already contains block with ID:%s" % block_id)
        else:
            return self._add_block(block_id).setParent(self)

    def get_next_id(self):
        return self._get_next_id()


    def get_k(self):
        return self._getK()

    @property
    def k(self):
        return self._getK()

    def free(self):
        self._free()


    def scan_layer(self, layer):
        self._scan_layer(layer)


    def insert_block_content(self, block):
        self._insert_block_content(block)

    def export_keyword(self, kw):
        if len(kw) != self.grid_ref.getGlobalSize():
            msg = 'The size of the target keyword must be equal to the size of the grid.'
            msg += '  Got:%d Expected:%d.'
            raise ValueError(msg % (len(kw), self.grid_ref.getGlobalSize()))

        if not kw.data_type.is_int():
            raise TypeError("The target kewyord must be of integer type")

        self._export_kw(kw)


    def add_fault_barrier(self, fault, link_segments=False):
        layer = self.getGeoLayer()
        layer.addFaultBarrier(fault, self.getK(), link_segments)


    def add_fault_link(self, fault1, fault2):
        if not fault1.intersectsFault(fault2, self.getK()):
            layer = self.getGeoLayer()
            layer.addIJBarrier(fault1.extendToFault(fault2, self.getK()))


    def join_faults(self, fault1, fault2):
        if not fault1.intersectsFault(fault2, self.getK()):
            layer = self.getGeoLayer()
            try:
                layer.addIJBarrier(Fault.joinFaults(fault1, fault2, self.getK()))
            except ValueError:
                err = 'Failed to join faults %s and %s'
                names = (fault1.getName(), fault2.getName())
                print(err % names)
                raise ValueError(err % names)


    def add_polyline_barrier(self, polyline):
        layer = self.getGeoLayer()
        p0 = polyline[0]
        c0 = self.grid_ref.findCellCornerXY(p0[0],  p0[1], self.getK())
        i,j = self.grid_ref.findCellXY(p0[0],  p0[1], self.getK())
        print('%g,%g -> %d,%d   %d' % (p0[0], p0[1], i,j,c0))
        for index in range(1,len(polyline)):
            p1 = polyline[index]
            c1 = self.grid_ref.findCellCornerXY(p1[0],  p1[1], self.getK())
            i,j = self.grid_ref.findCellXY(p1[0],  p1[1], self.getK())
            layer.addInterpBarrier(c0, c1)
            print('%g,%g -> %d,%d   %d' % (p1[0], p1[1], i,j,c1))
            print('Adding barrier %d -> %d' % (c0, c1))
            c0 = c1


    def get_geo_layer(self):
        """Returns the underlying geometric layer."""
        return self._get_layer()


    def cell_contact(self, p1, p2):
        layer = self.getGeoLayer()
        return layer.cellContact(p1,p2)

monkey_the_camel(FaultBlockLayer, 'scanKeyword', FaultBlockLayer.scan_keyword)
monkey_the_camel(FaultBlockLayer, 'loadKeyword', FaultBlockLayer.load_keyword)
monkey_the_camel(FaultBlockLayer, 'getBlock', FaultBlockLayer.get_block)
monkey_the_camel(FaultBlockLayer, 'deleteBlock', FaultBlockLayer.delete_block)
monkey_the_camel(FaultBlockLayer, 'addBlock', FaultBlockLayer.add_block)
monkey_the_camel(FaultBlockLayer, 'getNextID', FaultBlockLayer.get_next_id)
monkey_the_camel(FaultBlockLayer, 'getK', FaultBlockLayer.get_k)
monkey_the_camel(FaultBlockLayer, 'scanLayer', FaultBlockLayer.scan_layer)
monkey_the_camel(FaultBlockLayer, 'insertBlockContent', FaultBlockLayer.insert_block_content)
monkey_the_camel(FaultBlockLayer, 'exportKeyword', FaultBlockLayer.export_keyword)
monkey_the_camel(FaultBlockLayer, 'addFaultBarrier', FaultBlockLayer.add_fault_barrier)
monkey_the_camel(FaultBlockLayer, 'addFaultLink', FaultBlockLayer.add_fault_link)
monkey_the_camel(FaultBlockLayer, 'joinFaults', FaultBlockLayer.join_faults)
monkey_the_camel(FaultBlockLayer, 'addPolylineBarrier', FaultBlockLayer.add_polyline_barrier)
monkey_the_camel(FaultBlockLayer, 'getGeoLayer', FaultBlockLayer.get_geo_layer)
monkey_the_camel(FaultBlockLayer, 'cellContact', FaultBlockLayer.cell_contact)



import collections
from .geometry_tools import GeometryTools

class Polyline(object):
    def __init__(self, name=None , init_points = None):
        super(Polyline, self).__init__()
        self.__name = name
        self.__points = []
        if init_points:
            self.loadPoints( init_points )

    def __str__(self):
        s = "Polyline:[ "
        for p in self:
            s += "(%s,%s) " % (p[0],p[1])
        s += "]"
        return s

    def getName(self):
        """ @rtype: str """
        return self.__name


    def __iadd__(self , other ):
        for p in other:
            self.__points.append( p )
        return self


    def __add__(self , other ):
        copy = Polyline( init_points = self)
        copy.__iadd__(other)
        return copy


    def __radd__(self , other ):
        copy = Polyline( init_points = other )
        copy.__iadd__(self)
        return copy
            

    def __eq__(self, other):
        if len(self) != len(other):
            return False

        for (p1,p2) in zip(self , other):
            if p1 != p2:
                return False
        
        return True



    def __len__(self):
        return len(self.__points)

    def addPoint(self, x, y, z=None):
        if z is None:
            p = (x, y)
        else:
            p = (x, y, z)
        self.__points.append(p)

    def __getitem__(self, index):
        """ @rtype: tuple of (float, float, float) """
        if index < 0:
            index += len(self)

        if not 0 <= index < len(self):
            raise IndexError("Index:%d invalid must be in range: [0, %d]" % (index, (len(self) - 1)))
        return self.__points[index]

    def isClosed(self):
        first = self[0]
        last = self[-1]

        return first == last


    def assertClosed(self):
        if not self.isClosed():
            first_point = self[0]

            x = first_point[0]
            y = first_point[1]
            try:
                z = first_point[2]
                self.addPoint(x,y,z)
            except IndexError:
                self.addPoint(x,y)
            


            

    def loadPoints(self , points):
        for point in points:
            x = point[0]
            y = point[1]
            try:
                z = point[2]
            except IndexError:
                z = None
            
            self.addPoint(x,y,z)
                
    def intersects(self, other_polyline):
        """
        Test if instances intersects with other polyline.

        @type other_polyline: Polyline or list of tuple of (float, float)
        @rtype: bool
        """
        return GeometryTools.polylinesIntersect( self , other_polyline )


    def __iter__(self):
        index = 0

        while index < len(self):
            yield self[index]
            index += 1


    def unzip2(self):
        x = []
        y = []
        for p in self:
            x.append(p[0])
            y.append(p[1])
        
        return (x,y)


    def unzip(self):
        first_point = self[0]
        x = []
        y = []

        try:
            z = first_point[2]
            z = []
            for p in self:
                x.append(p[0])
                y.append(p[1])
                z.append(p[2])

            return (x,y,z)
        except IndexError:
            for p in self:
                x.append(p[0])
                y.append(p[1])
            
            return (x,y)
            

    def connect(self , target):
        end1 = self[0]
        end2 = self[-1]

        p1 = GeometryTools.nearestPointOnPolyline( end1 , target )
        p2 = GeometryTools.nearestPointOnPolyline( end2 , target )
            
        d1 = GeometryTools.distance( p1 , end1 )
        d2 = GeometryTools.distance( p2 , end2 )

        if d1 < d2:
            return [end1 , p1]
        else:
            return [end2 , p2]




#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_rft.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Module for loading ECLIPSE RFT files.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from cwrap import BaseCClass

from ecl import EclPrototype
from ecl.util.util import monkey_the_camel
from ecl.util.util import CTime
from ecl.rft import EclRFTCell, EclPLTCell

class EclRFT(BaseCClass):
    """The EclRFT class contains the information for *one* RFT.

    The ECLIPSE RFT file can contain three different types of RFT like
    objects which are lumped together; the EclRFTClass is a container
    for such objects. The three different object types which can be
    found in an RFT file are:

       RFT: This is old-fashioned RFT which contains measurements of
            saturations for each of the completed cells.

       PLT: This contains production and flow rates for each phase in
            each cell.

       SEGMENT: Not implemented.

    In addition to the measurements specific for RFT and PLT each cell
    has coordinates, pressure and depth.
    """
    TYPE_NAME = "ecl_rft"
    _alloc            = EclPrototype("void* ecl_rft_node_alloc_new( char* , char* , time_t , double)" , bind = False)
    _free             = EclPrototype("void  ecl_rft_node_free( ecl_rft )")
    _get_type         = EclPrototype("int    ecl_rft_node_get_type( ecl_rft )")
    _get_well         = EclPrototype("char*  ecl_rft_node_get_well_name( ecl_rft )")
    _get_date         = EclPrototype("time_t ecl_rft_node_get_date( ecl_rft )")
    _get_size         = EclPrototype("int ecl_rft_node_get_size( ecl_rft )")
    _iget_cell        = EclPrototype("void* ecl_rft_node_iget_cell( ecl_rft )")
    _iget_cell_sorted = EclPrototype("void* ecl_rft_node_iget_cell_sorted( ecl_rft )")
    _sort_cells       = EclPrototype("void* ecl_rft_node_inplace_sort_cells( ecl_rft )")
    _iget_depth       = EclPrototype("double ecl_rft_node_iget_depth( ecl_rft )")
    _iget_pressure    = EclPrototype("double ecl_rft_node_iget_pressure(ecl_rft)")
    _iget_ijk         = EclPrototype("void ecl_rft_node_iget_ijk( ecl_rft , int , int*, int*, int*)")
    _iget_swat        = EclPrototype("double ecl_rft_node_iget_swat(ecl_rft)")
    _iget_sgas        = EclPrototype("double ecl_rft_node_iget_sgas(ecl_rft)")
    _iget_orat        = EclPrototype("double ecl_rft_node_iget_orat(ecl_rft)")
    _iget_wrat        = EclPrototype("double ecl_rft_node_iget_wrat(ecl_rft)")
    _iget_grat        = EclPrototype("double ecl_rft_node_iget_grat(ecl_rft)")
    _lookup_ijk       = EclPrototype("void* ecl_rft_node_lookup_ijk( ecl_rft , int , int , int)")
    _is_RFT           = EclPrototype("bool   ecl_rft_node_is_RFT( ecl_rft )")
    _is_PLT           = EclPrototype("bool   ecl_rft_node_is_PLT( ecl_rft )")
    _is_SEGMENT       = EclPrototype("bool   ecl_rft_node_is_SEGMENT( ecl_rft )")
    _is_MSW           = EclPrototype("bool   ecl_rft_node_is_MSW( ecl_rft )")


    def __init__(self , name , type_string , date , days):
        c_ptr = self._alloc( name , type_string , CTime( date ) , days )
        super(EclRFT , self).__init__( c_ptr )


    def free(self):
        self._free( )

    def __repr__(self):
        rs = []
        rs.append('completed_cells = %d' % len(self))
        rs.append('date = %s' % self.getDate())
        if self.is_RFT():
            rs.append('RFT')
        if self.is_PLT():
            rs.append('PLT')
        if self.is_SEGMENT():
            rs.append('SEGMENT')
        if self.is_MSW():
            rs.append('MSW')
        rstr = ', '.join(rs)
        return self._create_repr(rstr)

    def __len__(self):
        """
        The number of completed cells in this RFT.
        """
        return self._get_size( )

    def is_RFT(self):
        """
        Is instance an RFT; in that case all the cells will be EclRFTCell instances.
        """
        return self._is_RFT( )

    def is_PLT(self):
        """
        Is instance a PLT; in that case all the cells will be EclPLTCell instances.
        """
        return self._is_PLT( )

    def is_SEGMENT(self):
        """
        Is this a SEGMENT - not implemented.
        """
        return self._is_SEGMENT( )

    def is_MSW(self):
        """
        Is this well a MSW well. Observe that the test ONLY applies to PLTs.
        """
        return self._is_MSW( )


    def get_well_name(self):
        """
        The name of the well we are considering.
        """
        return self._get_well( )

    def get_date(self):
        """
        The date when this RFT/PLT/... was recorded.
        """
        ct = CTime(self._get_date( ))
        return ct.date()

    def __cell_ref( self , cell_ptr ):
        if self.is_RFT():
            return EclRFTCell.createCReference( cell_ptr , self )
        elif self.is_PLT():
            return EclPLTCell.createCReference( cell_ptr , self )
        else:
            raise NotImplementedError("Only RFT and PLT cells are implemented")


    def assert_cell_index( self , index ):
        if isinstance( index , int):
            length = self.__len__()
            if index < 0 or index >= length:
                raise IndexError
        else:
            raise TypeError("Index should be integer type")


    def __getitem__(self , index):
        """Implements the [] operator to return the cells.

        To get the object related to cell nr 5:

           cell = rft[4]

        The return value from the __getitem__() method is either an
        EclRFTCell instance or a EclPLTCell instance, depending on the
        type of this particular RFT object.
        """
        self.assert_cell_index( index )
        cell_ptr = self._iget_cell( index )
        return self.__cell_ref( cell_ptr )


    def iget( self , index ):
        return self[index]


    def iget_sorted( self , index ):
        """
        Will return the cell nr @index in the list of sorted cells.

        See method sort() for further documentation.
        """
        self.assert_cell_index( index )
        cell_ptr = self._iget_cell_sorted( index )
        return self.__cell_ref( cell_ptr )


    def sort(self):
        """
        Will sort cells in RFT; currently only applies to MSW wells.

        By default the cells in the RFT come in the order they are
        specified in the ECLIPSE input file; that is not necessarily
        in a suitable order. In the case of MSW wells it is possible
        to sort the connections after distance along the wellpath. To
        access the cells in sort order you have two options:

           1. Sort the cells using the sort() method, and then
              subsequently access them sequentially:

                rft.sort()
                for cell in rft:
                    print cell

           2. Let the rft object stay unsorted, but access the cells
              using the iget_sorted() method:

                 for i in range(len(rft)):
                     cell = rft.iget_sorted( i )

        Currently only MSW/PLTs are sorted, based on the CONLENST
        keyword; for other wells the sort() method does nothing.
        """
        self._sort_cells( )


    # ijk are zero offset
    def ijkget( self , ijk ):
        """
        Will look up the cell based on (i,j,k).

        If the cell (i,j,k) is not part of this RFT/PLT None will be
        returned. The (i,j,k) input values should be zero offset,
        i.e. you must subtract 1 from the (i,j,k) values given in the ECLIPSE input.
        """
        cell_ptr = self._lookup_ijk( ijk[0] , ijk[1] , ijk[2])
        if cell_ptr:
            return self.__cell_ref( cell_ptr )
        else:
            return None




class EclRFTFile(BaseCClass):
    TYPE_NAME = "ecl_rft_file"
    _load          = EclPrototype("void* ecl_rft_file_alloc_case( char* )", bind = False)
    _iget          = EclPrototype("ecl_rft_ref ecl_rft_file_iget_node( ecl_rft_file , int )")
    _get_rft       = EclPrototype("ecl_rft_ref ecl_rft_file_get_well_time_rft( ecl_rft_file , char* , time_t)")
    _has_rft       = EclPrototype("bool ecl_rft_file_case_has_rft( char* )", bind = False)
    _free          = EclPrototype("void ecl_rft_file_free( ecl_rft_file )")
    _get_size      = EclPrototype("int ecl_rft_file_get_size__( ecl_rft_file , char* , time_t)")
    _get_num_wells = EclPrototype("int  ecl_rft_file_get_num_wells( ecl_rft_file )")


    """
    The EclRFTFile class is used to load an ECLIPSE RFT file.

    The EclRFTFile serves as a container which can load and hold the
    content of an ECLIPSE RFT file. The RFT files will in general
    contain data for several wells and several times in one large
    container. The EclRFTClass class contains methods get the the RFT
    results for a specific time and/or well.

    The EclRFTFile class can in general contain a mix of RFT and PLT
    measurements. The class does not really differentiate between
    these.
    """

    def __init__(self , case):
        c_ptr = self._load( case )
        super(EclRFTFile , self).__init__(c_ptr)


    def __len__(self):
        return self._get_size(  None , CTime(-1))


    def __getitem__(self, index):
        if isinstance(index, int):
            if 0 <= index < len(self):
                rft = self._iget(index)
                rft.setParent( self )
                return rft
            else:
                raise IndexError("Index '%d' must be in range: [0, %d]" % (index, len(self) - 1))
        else:
            raise TypeError("Index must be integer type")


    def size(self, well=None, date=None):
        """
        The number of elements in EclRFTFile container.

        By default the size() method will return the total number of
        RFTs/PLTs in the container, but by specifying the optional
        arguments date and/or well the function will only count the
        number of well measurements matching that time or well
        name. The well argument can contain wildcards.

           rftFile = ecl.EclRFTFile( "ECLIPSE.RFT" )
           print "Total number of RFTs : %d" % rftFile.size( )
           print "RFTs matching OP*    : %d" % rftFile.size( well = "OP*" )
           print "RFTs at 01/01/2010   : %d" % rftFile.size( date = datetime.date( 2010 , 1 , 1 ))

        """
        if date:
            cdate = CTime( date )
        else:
            cdate = CTime( -1 )

        return self._get_size( well , cdate)


    def get_num_wells(self):
        """
        Returns the total number of distinct wells in the RFT file.
        """
        return self._get_num_wells( )


    def get_headers(self):
        """
        Returns a list of two tuples (well_name , date) for the whole file.
        """
        header_list = []
        for i in (range(self._get_size( None , CTime(-1)))):
            rft = self.iget( i )
            header_list.append( (rft.getWellName() , rft.getDate()) )
        return header_list


    def iget(self , index):
        """
        Will lookup RFT @index - equivalent to [@index].
        """
        return self[index]


    def get(self , well_name , date ):
        """
        Will look up the RFT object corresponding to @well and @date.

        Raise Exception if not found.
        """
        if self.size( well = well_name , date = date) == 0:
            raise KeyError("No RFT for well:%s at %s" % (well_name , date))

        rft = self._get_rft( well_name , CTime( date ))
        rft.setParent( self )
        return rft

    def free(self):
        self._free( )

    def __repr__(self):
        w = len(self)
        return self._create_repr('wells = %d' % w)


monkey_the_camel(EclRFT, 'getWellName', EclRFT.get_well_name)
monkey_the_camel(EclRFT, 'getDate', EclRFT.get_date)

monkey_the_camel(EclRFTFile, 'getNumWells', EclRFTFile.get_num_wells)
monkey_the_camel(EclRFTFile, 'getHeaders', EclRFTFile.get_headers)



import os
from .polyline import Polyline

class XYZIo(object):

    @staticmethod
    def readXYZFile(path):
        """ @rtype: Polyline """

        if not os.path.exists(path):
            raise IOError("Path does not exist '%s'!" % path)

        name = os.path.basename(path)

        polyline = Polyline(name=name)

        with open(path, "r") as f:
            for line in f:
                line = line.strip()
                if line:
                    x, y, z = map(float, line.split())
                
                    if x != 999.000000 and y != 999.000000 and z != 999.000000:
                        polyline.addPoint(x, y, z)
                    else:
                        break
                else:
                    break

        return polyline

        
    @staticmethod
    def readXYFile(path):
        """ @rtype: Polyline """

        if not os.path.exists(path):
            raise IOError("Path does not exist '%s'!" % path)

        name = os.path.basename(path)

        polyline = Polyline(name=name)

        with open(path, "r") as f:
            for line in f:
                x, y= map(float, line.split())
                polyline.addPoint(x, y)

        return polyline


    @staticmethod
    def saveXYFile(polyline , filename):
        """
        @type polyline: Polyline or list of tuple of (float, float)
        """
        with open(filename , "w") as fileH:
            for p in polyline:
                fileH.write("%g %g\n" % (p[0] , p[1]))
        





#  Copyright (C) 2015  Statoil ASA, Norway.
#
#  The file 'ecl_restart_file.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from cwrap import BaseCClass

from ecl import EclPrototype
from ecl.util.util import monkey_the_camel
from ecl.util.util import CTime
from ecl import EclFileEnum
from ecl.eclfile import EclFile, Ecl3DFile, Ecl3DKW

class EclRestartHead(BaseCClass):
    TYPE_NAME = "ecl_rsthead"
    _alloc           = EclPrototype("void*  ecl_rsthead_alloc(ecl_file_view , int )", bind = False)
    _alloc_from_kw   = EclPrototype("void*  ecl_rsthead_alloc_from_kw(int , ecl_kw , ecl_kw , ecl_kw )", bind = False)
    _free            = EclPrototype("void   ecl_rsthead_free(ecl_rsthead)")
    _get_report_step = EclPrototype("int    ecl_rsthead_get_report_step(ecl_rsthead)")
    _get_sim_time    = EclPrototype("time_t ecl_rsthead_get_sim_time(ecl_rsthead)")
    _get_sim_days    = EclPrototype("double ecl_rsthead_get_sim_days(ecl_rsthead)")
    _get_nxconz      = EclPrototype("int   ecl_rsthead_get_nxconz(ecl_rsthead)")
    _get_ncwmax      = EclPrototype("int   ecl_rsthead_get_ncwmax(ecl_rsthead)")

    def __init__(self , kw_arg = None , rst_view = None):
        if kw_arg is None and rst_view is None:
            raise ValueError('Cannot construct EclRestartHead without one of kw_arg and rst_view, both were None!')

        if not kw_arg is None:
            report_step , intehead_kw , doubhead_kw , logihead_kw = kw_arg
            c_ptr = self._alloc_from_kw( report_step , intehead_kw , doubhead_kw , logihead_kw )
        else:
            c_ptr = self._alloc( rst_view , -1 )

        super(EclRestartHead, self).__init__(c_ptr)


    def free(self):
        self._free( )

    def get_report_step(self):
        return self._get_report_step( )

    def get_sim_date(self):
        ct = CTime( self._get_sim_time( ) )
        return ct.datetime( )

    def get_sim_days(self):
        return self._get_sim_days( )

    def well_details(self):
        return {"NXCONZ" : self._get_nxconz(),
                "NCWMAX" : self._get_ncwmax()}




class EclRestartFile(Ecl3DFile):

    def __init__(self , grid , filename , flags = 0):
        """Will open an Eclipse restart file.

        The EclRestartFile class will open an eclipse restart file, in
        unified or non unified format. The constructor will infer the
        file type based on the filename, and will raise a ValueError
        exception if the file type is not ECL_RESTART_FILE or
        ECL_UNIFIED_RESTART_FILE.

        The EclRestartFile will use a grid reference to create Ecl3DKw
        instances for all the keyword elements which have either
        'nactive' or 'nx*ny*nz' elements.
        """

        file_type , report_step , fmt_file = EclFile.getFileType( filename )
        if not file_type in [EclFileEnum.ECL_RESTART_FILE, EclFileEnum.ECL_UNIFIED_RESTART_FILE]:
            raise ValueError('The input filename "%s" does not correspond to a restart file.  Please follow the Eclipse naming conventions'
                             % filename)

        super(EclRestartFile , self).__init__( grid, filename , flags)
        self.rst_headers = None
        if file_type == EclFileEnum.ECL_RESTART_FILE:
            self.is_unified = False
            self.report_step = report_step
        else:
            self.is_unified = True



    def unified(self):
        """
        Will return True if the file we have opened is unified.
        """
        return self.is_unified


    def assert_headers(self):
        if self.rst_headers is None:
            self.rst_headers = []
            if self.unified():
                for index in range(self.num_named_kw("SEQNUM")):
                    self.rst_headers.append( EclRestartHead( rst_view = self.restartView( seqnum_index = index )))
            else:
                intehead_kw = self["INTEHEAD"][0]
                doubhead_kw = self["DOUBHEAD"][0]
                if "LOGIHEAD" in self:
                    logihead_kw = self["LOGIHEAD"][0]
                else:
                    logihead_kw = None

                self.rst_headers.append( EclRestartHead( kw_arg = (self.report_step , intehead_kw , doubhead_kw , logihead_kw) ))


    def time_list(self):
        """Will return a list of report_step, simulation time and days.

        The return value will be a list tuples. For a unified restart
        file with the three report steps {10,15,20} it can look like:

           [  (10, datetime.datetime( 2010 , 1 , 1 , 0 , 0 , 0 ) , 100.0),
              (15, datetime.datetime( 2010 , 3 , 1 , 0 , 0 , 0 ) , 160.0),
              (20, datetime.datetime( 2010 , 5 , 1 , 0 , 0 , 0 ) , 220.0) ]

        For a non-unified restart file the list will have only one element.
        """

        self.assertHeaders()
        time_list = []
        for header in self.rst_headers:
            time_list.append( (header.getReportStep() , header.getSimDate( ) , header.getSimDays( )) )

        return time_list


    def headers(self):
        self.assertHeaders()
        return self.rst_headers


    def get_header(self, index):
        self.assertHeaders()
        return self.rst_headers[index]

monkey_the_camel(EclRestartHead, 'getReportStep', EclRestartHead.get_report_step)
monkey_the_camel(EclRestartHead, 'getSimDate', EclRestartHead.get_sim_date)
monkey_the_camel(EclRestartHead, 'getSimDays', EclRestartHead.get_sim_days)

monkey_the_camel(EclRestartFile, 'assertHeaders', EclRestartFile.assert_headers)
monkey_the_camel(EclRestartFile, 'timeList', EclRestartFile.time_list)



from .test_run import TestRun
from .test_run import path_exists
from .extended_testcase import ExtendedTestCase
from .source_enumerator import SourceEnumerator
from .test_area import TestArea , TestAreaContext
from .temp_area import TempArea , TempAreaContext
from .ert_test_runner import ErtTestRunner
from .path_context import PathContext
from .lint_test_case import LintTestCase
from .import_test_case import ImportTestCase
from .debug_msg import debug_msg



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'fortio.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Module to support transparent binary IO of Fortran created files.

Fortran is a a funny language; when writing binary blobs of data to
file the Fortran runtime will silently add a header and footer around
the date. The Fortran code:

   integer array(100)
   write(unit) array

it actually writes a head and tail in addition to the actual
data. The header and tail is a 4 byte integer, which value is the
number of bytes in the immediately following record. I.e. what is
actually found on disk after the Fortran code above is:

  | 400 | array ...... | 400 |

The fortio.c file implements the fortio_type C structure which can be
used to read and write these structures transparently. The current
python module is a minimal wrapping of this datastructure; mainly to
support passing of FortIO handles to the underlying C functions. A
more extensive wrapping of the fortio implementation would be easy.
"""
import ctypes
import os

from cwrap import BaseCClass
from ecl.util.util import monkey_the_camel
from ecl import EclPrototype


class FortIO(BaseCClass):
    TYPE_NAME = "fortio"

    READ_MODE = 1
    WRITE_MODE = 2
    READ_AND_WRITE_MODE = 3
    APPEND_MODE = 4

    _open_reader    = EclPrototype("void* fortio_open_reader(char*, bool, bool)", bind=False)
    _open_writer    = EclPrototype("void* fortio_open_writer(char*, bool, bool)", bind=False)
    _open_readwrite = EclPrototype("void* fortio_open_readwrite(char*, bool, bool)", bind=False)
    _open_append    = EclPrototype("void* fortio_open_append(char*, bool, bool)", bind=False)
    _guess_fortran  = EclPrototype("bool fortio_looks_like_fortran_file(char*, bool)", bind=False)

    _write_record   = EclPrototype("void fortio_fwrite_record(fortio, char*, int)")
    _get_position   = EclPrototype("long fortio_ftell(fortio)")
    _seek           = EclPrototype("void fortio_fseek(fortio, long, int)")
    _close          = EclPrototype("bool fortio_fclose(fortio)")
    _truncate       = EclPrototype("bool fortio_ftruncate(fortio, long)")
    _filename       = EclPrototype("char* fortio_filename_ref(fortio)")


    def __init__(self, file_name, mode=READ_MODE, fmt_file=False, endian_flip_header=True):
        """Will open a new FortIO handle to @file_name - default for reading.

        The newly created FortIO handle will open the underlying FILE*
        for reading, but if you pass the flag mode=FortIO.WRITE_MODE
        the file will be opened for writing.

        Observe that the flag @endian_flip_header will only affect the
        interpretation of the block size markers in the file, endian
        flipping of the actual data blocks must be handled at a higher
        level.

        When you are finished working with the FortIO instance you can
        manually close it with the close() method, alternatively that
        will happen automagically when it goes out of scope.

        Small example script opening a restart file, and then writing
        all the pressure keywords to another file:

           import sys
           from ecl.ecl import FortIO, EclFile

           rst_file = EclFile(sys.argv[1])
           fortio = FortIO("PRESSURE", mode=FortIO.WRITE_MODE)

           for kw in rst_file:
               if kw.name() == "PRESSURE":
                  kw.write(fortio)

           fortio.close()

        See the documentation of openFortIO() for an alternative
        method based on a context manager and the with statement.

        """
        read_modes = (FortIO.READ_MODE, FortIO.APPEND_MODE, FortIO.READ_AND_WRITE_MODE)
        if mode in read_modes and not os.path.exists(file_name):
            raise IOError('No such file "%s".' % file_name)
        if mode == FortIO.READ_MODE:
            c_pointer = self._open_reader(file_name, fmt_file, endian_flip_header)
        elif mode == FortIO.WRITE_MODE:
            c_pointer = self._open_writer(file_name, fmt_file, endian_flip_header)
        elif mode == FortIO.READ_AND_WRITE_MODE:
            c_pointer = self._open_readwrite(file_name, fmt_file, endian_flip_header)
        elif mode == FortIO.APPEND_MODE:
            c_pointer = self._open_append(file_name, fmt_file, endian_flip_header)
        else:
            raise UserWarning("Unknown mode: %d" % mode)

        self.__mode = mode
        if not c_pointer:
            raise IOError('Failed to open FortIO file "%s".' % file_name)
        super(FortIO, self).__init__(c_pointer)



    def close(self):
        if self:
            self._close()
            self._invalidateCPointer()


    def get_position(self):
        """ @rtype: long """
        return self._get_position()


    def truncate(self, size=None):
        """Will truncate the file to new size.

        If the method is called without a size argument the stream
        will be truncated to the current position.
        """
        if size is None:
            size = self.getPosition()

        if not self._truncate(size):
            raise IOError("Truncate of fortran filehandle:%s failed" % self.filename())


    def filename(self):
        return self._filename()


    def seek(self, position, whence=0):
        # SEEK_SET = 0
        # SEEK_CUR = 1
        # SEEK_END = 2
        self._seek(position, whence)


    @classmethod
    def is_fortran_file(cls, filename, endian_flip=True):

        """@rtype: bool
        @type filename: str


        Will use heuristics to try to guess if @filename is a binary
        file written in fortran style. ASCII files will return false,
        even if they are structured as ECLIPSE keywords.
        """
        return cls._guess_fortran(filename, endian_flip)


    def free(self):
        self.close()


class FortIOContextManager(object):

    def __init__(self, fortio):
        self.__fortio = fortio

    def __enter__(self):
        return self.__fortio

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__fortio.close()
        return exc_type is not None


def openFortIO(file_name, mode=FortIO.READ_MODE, fmt_file=False, endian_flip_header=True):
    """Will create FortIO based context manager for use with with.

    The with: statement and context managers is a good alternative in
    the situation where you need to ensure resource cleanup.

       import sys
       from ecl.ecl import FortIO, EclFile

       rst_file = EclFile(sys.argv[1])
       with openFortIO("PRESSURE", mode=FortIO.WRITE_MODE) as fortio:
          for kw in rst_file:
              if kw.name() == "PRESSURE":
                 kw.write(fortio)

    """
    return FortIOContextManager(FortIO(file_name, mode=mode, fmt_file=fmt_file,
                                       endian_flip_header=endian_flip_header))

monkey_the_camel(FortIO, 'getPosition', FortIO.get_position)
monkey_the_camel(FortIO, 'isFortranFile', FortIO.is_fortran_file, classmethod)




#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'matrix.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.


# The Matrix class implemented here wraps the C matrix implementation
# in matrix.c from the libutil library. The C matrix implementation
# has the very limited ambition of just barely satisfying the matrix
# needs of the EnKF algorithm, i.e. for general linear algebra
# applications you will probably be better served by a more complete
# matrix library. This applies even more so to this Python
# implementation; it is only here facilitate use of C libraries which
# expect a matrix instance as input (i.e. the LARS estimator). For
# general linear algebra in Python the numpy library is a natural
# choice.


from cwrap import BaseCClass,CFILE
from ecl import EclPrototype


class Matrix(BaseCClass):
    _matrix_alloc      = EclPrototype("void*  matrix_alloc(int, int )" , bind = False)
    _matrix_alloc_identity = EclPrototype("matrix_obj  matrix_alloc_identity( int )" , bind = False)
    _alloc_transpose   = EclPrototype("matrix_obj  matrix_alloc_transpose(matrix)")
    _inplace_transpose = EclPrototype("void        matrix_inplace_transpose(matrix)")
    _copy              = EclPrototype("matrix_obj  matrix_alloc_copy(matrix)" )
    _sub_copy          = EclPrototype("matrix_obj  matrix_alloc_sub_copy(matrix, int , int , int , int)" )
    _free              = EclPrototype("void   matrix_free(matrix)")
    _iget              = EclPrototype("double matrix_iget( matrix , int , int )")
    _iset              = EclPrototype("void   matrix_iset( matrix , int , int , double)")
    _set_all           = EclPrototype("void   matrix_scalar_set( matrix , double)")
    _scale_column      = EclPrototype("void matrix_scale_column(matrix , int , double)")
    _scale_row         = EclPrototype("void matrix_scale_row(matrix , int , double)")
    _copy_column       = EclPrototype("void matrix_copy_column(matrix , matrix , int , int)" , bind = False)
    _rows              = EclPrototype("int matrix_get_rows(matrix)")
    _columns           = EclPrototype("int matrix_get_columns(matrix)")
    _equal             = EclPrototype("bool matrix_equal(matrix, matrix)")
    _pretty_print      = EclPrototype("void matrix_pretty_print(matrix, char*, char*)")
    _fprint            = EclPrototype("void matrix_fprintf(matrix, char*, FILE)")
    _random_init       = EclPrototype("void matrix_random_init(matrix, rng)")
    _dump_csv          = EclPrototype("void matrix_dump_csv(matrix, char*)")

    # Requires BLAS. If the library does not have the
    # matrix_alloc_matmul() function the prototype will have _func =
    # None, and NotImplementedError( ) will be raised int the
    # __call__() method if we try to use this function.
    try:
        _alloc_matmul = EclPrototype("matrix_obj  matrix_alloc_matmul(matrix, matrix)" , bind = False)
    except AttributeError:
        _alloc_matmul = None

    # Requires BLAS!
    @classmethod
    def matmul(cls, m1,m2):
        """
        Will return a new matrix which is matrix product of m1 and m2.
        """
        if m1.columns( ) == m2.rows( ):
            return cls._alloc_matmul( m1, m2)
        else:
            raise ValueError("Matrix size mismatch")


    def __init__(self, rows, columns, value=0):
        c_ptr = self._matrix_alloc(rows, columns)
        super(Matrix, self).__init__(c_ptr)
        self.setAll(value)

    def copy(self):
        return self._copy( )

    @classmethod
    def identity(cls, dim):
        """Returns a dim x dim identity matrix."""
        if dim < 1:
            raise ValueError('Identity matrix must have positive size, %d not allowed.' % dim)
        return cls._matrix_alloc_identity(dim)

    def subCopy(self, row_offset, column_offset, rows, columns):
        if row_offset < 0 or row_offset >= self.rows():
            raise ValueError("Invalid row offset")

        if column_offset < 0 or column_offset >= self.columns():
            raise ValueError("Invalid column offset")

        if row_offset + rows > self.rows():
            raise ValueError("Invalid rows")

        if column_offset + columns > self.columns():
            raise ValueError("Invalid columns")

        return self._sub_copy( row_offset , column_offset , rows , columns)


    def __str__(self):
        s = ""
        for i in range(self.rows()):
            s += "["
            for j in range(self.columns()):
                d = self._iget(i, j)
                s += "%6.3g " % d
            s += "]\n"
        return s

    def __getitem__(self, index_tuple):
        if not 0 <= index_tuple[0] < self.rows():
            raise IndexError("Expected 0 <= %d < %d" % (index_tuple[0], self.rows()))

        if not 0 <= index_tuple[1] < self.columns():
            raise IndexError("Expected 0 <= %d < %d" % (index_tuple[1], self.columns()))

        return self._iget(index_tuple[0], index_tuple[1])

    def __setitem__(self, index_tuple, value):
        if not 0 <= index_tuple[0] < self.rows():
            raise IndexError("Expected 0 <= %d < %d" % (index_tuple[0], self.rows()))

        if not 0 <= index_tuple[1] < self.columns():
            raise IndexError("Expected 0 <= %d < %d" % (index_tuple[1], self.columns()))

        return self._iset(index_tuple[0], index_tuple[1], value)

    def dims(self):
        return self._rows(), self._columns()

    def rows(self):
        """ @rtype: int """
        return self._rows()

    def transpose(self , inplace = False):
        """
        Will transpose the matrix. By default a transposed copy is returned.
        """
        if inplace:
            self._inplace_transpose( )
            return self
        else:
            return self._alloc_transpose( )


    def columns(self):
        """ @rtype: int """
        return self._columns()

    def __eq__(self, other):
        assert isinstance(other, Matrix)
        return self._equal(other)

    def scaleColumn(self, column, factor):
        if not 0 <= column < self.columns():
            raise IndexError("Expected column: [0,%d) got:%d" % (self.columns(), column))
        self._scale_column(column, factor)

    def scaleRow(self, row, factor):
        if not 0 <= row < self.rows():
            raise IndexError("Expected row: [0,%d) got:%d" % (self.rows(), row))
        self._scale_row(row, factor)

    def setAll(self, value):
        self._set_all(value)

    def copyColumn(self, target_column, src_column):
        columns = self.columns()
        if not 0 <= src_column < columns:
            raise ValueError("src column:%d invalid" % src_column)

        if not 0 <= target_column < columns:
            raise ValueError("target column:%d invalid" % target_column)

        if src_column != target_column:
            # The underlying C function accepts column copy between matrices.
            Matrix._copy_column(self, self, target_column, src_column)


    def dumpCSV(self , filename):
        self._dump_csv( filename )


    def prettyPrint(self, name, fmt="%6.3g"):
        self._pretty_print(name, fmt)

    def fprint(self , fileH , fmt = "%g "):
        """Will print ASCII representation of matrix.

        The fileH argument should point to an open Python
        filehandle. If you supply a fmt string it is important that it
        contains a separator, otherwise you might risk that elements
        overlap in the output. For the matrix:

                  [0 1 2]
              m = [3 4 5]
                  [6 7 8]

        The code:

        with open("matrix.txt" , "w") as f:
           m.fprintf( f )

         The file 'matrix.txt' will look like:

         0 1 2
         3 4 5
         6 7 8

        """
        self._fprint( fmt , CFILE( fileH))


    def randomInit(self, rng):
        self._random_init(rng)

    def free(self):
        self._free()






import multiprocessing
from threading import Thread
import time
import traceback


class Task(Thread):
    def __init__(self, func, *args, **kwargs):
        super(Task, self).__init__()
        self.__func = func
        self.__args = args
        self.__kwargs = kwargs

        self.__started = False
        self.__done = False
        self.__failed = False
        self.__start_time = None

        self.__verbose = False

    def setVerbose(self, verbose):
        self.__verbose = verbose

    def run(self):
        self.__start_time = time.time()
        self.__started = True
        try:
            self.__func(*self.__args, **self.__kwargs)
        except Exception:
            self.__failed = True
            traceback.print_exc()
        finally:
            self.__done = True

        if self.__verbose:
            running_time = time.time() - self.__start_time
            print("Running time of task: %f" % running_time)

    def isDone(self):
        return self.__done

    def hasStarted(self):
        return self.__started or self.isAlive()

    def isRunning(self):
        return self.hasStarted() and not self.__done

    def hasFailed(self):
        return self.__failed

    def join(self, timeout=None):
        while not self.hasStarted() or self.isRunning():
            time.sleep(0.01)



class ThreadPool(object):

    def __init__(self, size=None, verbose=False):
        super(ThreadPool, self).__init__()

        if size is None:
            size = multiprocessing.cpu_count()

        self.__size = size
        self.__task_list = []
        self.__pool_finished = False
        self.__runner_thread = None
        self.__verbose = verbose
        self.__start_time = None


    def addTask(self, func, *args, **kwargs):
        if self.__start_time is None:
            task = Task(func, *args, **kwargs)
            # task.setVerbose(self.__verbose)
            self.__task_list.append(task)
        else:
            raise UserWarning("Can not add task after the pool has started!")

    def poolSize(self):
        return self.__size

    def taskCount(self):
        return len(self.__task_list)

    def __allTasksFinished(self):
        for task in self.__task_list:
            if not task.isDone():
                return False

        return True

    def runningCount(self):
        count = 0

        for task in self.__task_list:
            if task.isRunning():
                count += 1

        return count

    def doneCount(self):
        count = 0

        for task in self.__task_list:
            if task.isDone():
                count += 1

        return count

    def __findNextTask(self):
        for task in self.__task_list:
            if not task.hasStarted():
                return task

        return None

    def __start(self):
        while not self.__allTasksFinished():
            if self.runningCount() < self.poolSize():
                task = self.__findNextTask()

                if task is not None:
                    task.start()

            time.sleep(0.001)

        self.__pool_finished = True


    def nonBlockingStart(self):
        self.__runner_thread = Thread()
        self.__runner_thread.run = self.__start
        self.__runner_thread.start()
        self.__start_time = time.time()

    def join(self):
        while not self.__pool_finished:
            time.sleep(0.001)

        if self.__verbose:
            running_time = time.time() - self.__start_time
            print("Running time: %f using a pool size of: %d" % (running_time, self.poolSize()))

    def hasFailedTasks(self):
        for task in self.__task_list:
            if task.hasFailed():
                return True

        return False



#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'fault_collection.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
import re

from ecl.util.util import monkey_the_camel
from ecl.grid import EclGrid
from .fault import Fault

comment_regexp = re.compile("--.*")

def dequote(s):
    if s[0] in ["'", '"']:
        if s[0] == s[-1]:
            return s[1:-1]
        else:
            raise ValueError("Quote fuckup")
    else:
        return s


class FaultCollection(object):
    def __init__(self, grid=None, *file_list):
        self.__fault_list = []
        self.__fault_map = {}
        self.__grid = grid

        if self.__grid is not None:
            if not isinstance(self.__grid, EclGrid):
                raise ValueError("When supplying a list of files to load - you must have a grid")
            for file in file_list:
                self.load(self.__grid, file)


    def __contains__(self, fault_name):
        return fault_name in self.__fault_map


    def __len__(self):
        return len(self.__fault_list)


    def __getitem__(self, index):
        if isinstance(index, str):
            return self.__fault_map[index]
        elif isinstance(index, int):
            return self.__fault_list[index]
        else:
            raise TypeError("Argument must be fault name or number")

    def __iter__(self):
        return iter(self.__fault_list)


    def get_grid(self):
        return self.__grid


    def get_fault(self, name):
        return self[name]


    def has_fault(self, fault_name):
        return fault_name in self


    def add_fault(self, fault):
        self.__fault_map[fault.getName()] = fault
        self.__fault_list.append(fault)


    def split_line(self, line):
        tmp = line.split()
        if not tmp[-1] == "/":
            raise ValueError("Line:%s does not end with /" % line)

        if len(tmp) != 9:
            raise ValueError("Line:%s not correct number of items" % line)

        fault_name = dequote(tmp[0])
        I1 = int(tmp[1]) - 1
        I2 = int(tmp[2]) - 1
        J1 = int(tmp[3]) - 1
        J2 = int(tmp[4]) - 1
        K1 = int(tmp[5]) - 1
        K2 = int(tmp[6]) - 1
        face = dequote(tmp[7])

        return (fault_name, I1,I2,J1,J2,K1,K2, face)



    def load_faults(self, grid, fileH):
        for line in fileH:
            line = comment_regexp.sub("", line)
            line = line.strip()
            if line == "/":
                break

            if line:
                (name, I1, I2, J1, J2, K1, K2, face) = self.splitLine(line)
                if not self.hasFault(name):
                    fault = Fault(grid, name)
                    self.addFault(fault)
                else:
                    fault = self.getFault(name)

                fault.addRecord(I1, I2, J1, J2, K1, K2, face)


    def load(self, grid, file_name):
        with open(file_name) as fileH:
            for line in fileH:
                if line.startswith("FAULTS"):
                    self.loadFaults(grid, fileH)


monkey_the_camel(FaultCollection, 'getGrid', FaultCollection.get_grid)
monkey_the_camel(FaultCollection, 'getFault', FaultCollection.get_fault)
monkey_the_camel(FaultCollection, 'hasFault', FaultCollection.has_fault)
monkey_the_camel(FaultCollection, 'addFault', FaultCollection.add_fault)
monkey_the_camel(FaultCollection, 'splitLine', FaultCollection.split_line)
monkey_the_camel(FaultCollection, 'loadFaults', FaultCollection.load_faults)



#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'fault_block_collection.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.


from cwrap import BaseCClass

from ecl.util.util import monkey_the_camel
from ecl import EclPrototype


class FaultBlockCollection(BaseCClass):
    TYPE_NAME = "fault_block_collection"
    _alloc         = EclPrototype("void* fault_block_collection_alloc(ecl_grid)", bind=False)
    _free          = EclPrototype("void  fault_block_collection_free(fault_block_collection)")
    _num_layers    = EclPrototype("int   fault_block_collection_num_layers(fault_block_collection)")
    _scan_keyword  = EclPrototype("bool  fault_block_collection_scan_kw(fault_block_collection, ecl_kw)")
    _get_layer     = EclPrototype("fault_block_layer_ref  fault_block_collection_get_layer(fault_block_collection, int)")

    def __init__(self, grid):
        c_ptr = self._alloc(grid)
        if c_ptr:
            super(FaultBlockCollection, self).__init__(c_ptr)
        else:
            raise ValueError("Invalid input - failed to create FaultBlockCollection")

        # The underlying C implementation uses lazy evaluation and
        # needs to hold on to the grid reference. We therefore take
        # references to it here, to protect against premature garbage
        # collection.
        self.grid_ref = grid


    def __len__(self):
        return self._num_layers()


    def __repr__(self):
        return self._create_repr('len=%s' % len(self))


    def __getitem__(self, index):
        """
        @rtype: FaultBlockLayer
        """
        if isinstance(index, int):
            if 0 <= index < len(self):
                return self._get_layer(index).setParent(self)
            else:
                raise IndexError("Index:%d out of range [0,%d)" % (index, len(self)))
        else:
            raise TypeError("Index should be integer type")


    def get_layer(self, k):
        """
        @rtype: FaultBlockLayer
        """
        return self[k]

    def free(self):
        self._free()


    def scan_keyword(self, fault_block_kw):
        ok = self._scan_keyword(fault_block_kw)
        if not ok:
            raise ValueError("The fault block keyword had wrong type/size")


monkey_the_camel(FaultBlockCollection, 'getLayer', FaultBlockCollection.get_layer)
monkey_the_camel(FaultBlockCollection, 'scanKeyword', FaultBlockCollection.scan_keyword)
